@@ -6,6 +6,8 @@
  */
 
 #include <fiwix/kernel.h>
+#include <fiwix/kparms.h>
+#include <fiwix/asm.h>
 #include <fiwix/ioctl.h>
 #include <fiwix/tty.h>
 #include <fiwix/ctype.h>
@@ -19,10 +21,17 @@
 #include <fiwix/process.h>
 #include <fiwix/fcntl.h>
 #include <fiwix/kd.h>
+#include <fiwix/pty.h>
+#include <fiwix/fs_devpts.h>
+#include <fiwix/mm.h>
 #include <fiwix/stdio.h>
 #include <fiwix/string.h>
 
-struct tty tty_table[NR_TTYS];
+#define SYSCON_DEV	MKDEV(SYSCON_MAJOR, 1)
+#define VCONSOLE_DEV	MKDEV(VCONSOLES_MAJOR, 0)
+#define TTY_DEV		MKDEV(SYSCON_MAJOR, 0)
+
+struct tty *tty_table;
 extern short int current_cons;
 
 static void wait_vtime_off(unsigned int arg)
@@ -89,6 +98,7 @@ static int opost(struct tty *tty, unsigned char ch)
 				if(tty->column > 0) {
 					tty->column--;
 				}
+				break;
 			default:
 				if(tty->termios.c_oflag & OLCUC) {
 					ch = TOUPPER(ch);
@@ -179,54 +189,108 @@ static void set_termio(struct tty *tty, struct termio *new_termio)
 	memcpy_b(&tty->termios, &new_termios, sizeof(struct termios));
 }
 
-int register_tty(__dev_t dev)
+void tty_reset(struct tty *tty)
 {
+	termios_reset(tty);
+	tty->winsize.ws_row = 25;
+	tty->winsize.ws_col = 80;
+	tty->winsize.ws_xpixel = 0;
+	tty->winsize.ws_ypixel = 0;
+	tty->flags = 0;
+}
+
+struct tty *register_tty(__dev_t dev)
+{
+	unsigned int flags;
+	struct tty *tty, *t;
 	int n;
 
-	for(n = 0; n < NR_TTYS; n++) {
-		if(tty_table[n].dev == dev) {
-			printk("ERROR: %s(): tty device %d,%d already registered!\n", __FUNCTION__, MAJOR(dev), MINOR(dev));
-			return 1;
+	if(!(tty = (struct tty *)kmalloc(sizeof(struct tty)))) {
+		return NULL;
+	}
+
+	SAVE_FLAGS(flags); CLI();
+	if((t = tty_table)) {
+		for(;;) {
+			if(t->dev == dev) {
+				printk("ERROR: %s(): tty device %d,%d already registered!\n", __FUNCTION__, MAJOR(dev), MINOR(dev));
+				RESTORE_FLAGS(flags);
+				kfree((unsigned int)tty);
+				return NULL;
+			}
+			if(t->next) {
+				t = t->next;
+			} else {
+				break;
+			}
 		}
-		if(!tty_table[n].dev) {
-			tty_table[n].dev = dev;
-			tty_table[n].count = 0;
-			return 0;
+		t->next = tty;
+	} else {
+		tty_table = tty;
+	}
+	RESTORE_FLAGS(flags);
+
+	memset_b(tty, 0, sizeof(struct tty));
+	tty_reset(tty);
+	tty->dev = dev;
+	for(n = 0; n < MAX_TAB_COLS; n++) {
+		if(!(n % TAB_SIZE)) {
+			tty->tab_stop[n] = 1;
+		} else {
+			tty->tab_stop[n] = 0;
 		}
 	}
-	printk("ERROR: %s(): tty table is full!\n", __FUNCTION__);
-	return 1;
+	return tty;
+}
+
+void unregister_tty(struct tty *tty)
+{
+	unsigned int flags;
+	struct tty *t;
+
+	SAVE_FLAGS(flags); CLI();
+	if(tty == tty_table) {
+		tty_table = tty->next;
+	} else {
+		t = tty_table;
+		while(t->next != tty) {
+			t = t->next;
+		}
+		t->next = tty->next;
+	}
+	kfree((unsigned int)tty);
+	RESTORE_FLAGS(flags);
 }
 
 struct tty *get_tty(__dev_t dev)
 {
-	int n;
+	struct tty *tty;
 
 	if(!dev) {
 		return NULL;
 	}
 
 	/* /dev/console = system console */
-	if(dev == MKDEV(SYSCON_MAJOR, 1)) {
-		dev = (__dev_t)kparm_syscondev;
+	if(dev == SYSCON_DEV) {
+		dev = (__dev_t)kparms.syscondev;
 	}
 
 	/* /dev/tty0 = current virtual console */
-	if(dev == MKDEV(VCONSOLES_MAJOR, 0)) {
+	if(dev == VCONSOLE_DEV) {
 		dev = MKDEV(VCONSOLES_MAJOR, current_cons);
 	}
 
 	/* /dev/tty = controlling TTY device */
-	if(dev == MKDEV(SYSCON_MAJOR, 0)) {
+	if(dev == TTY_DEV) {
 		if(!current->ctty) {
 			return NULL;
 		}
 		dev = current->ctty->dev;
 	}
 
-	for(n = 0; n < NR_TTYS; n++) {
-		if(tty_table[n].dev == dev) {
-			return &tty_table[n];
+	for(tty = tty_table; tty; tty = tty->next) {
+		if(tty->dev == dev) {
+			return tty;
 		}
 	}
 	return NULL;
@@ -281,6 +345,40 @@ void termios_reset(struct tty *tty)
 	tty->termios.c_cc[VEOL2]    = 0;	
 }
 
+void tty_deltab(struct tty *tty)
+{
+	unsigned short int col, n, count;
+	struct cblock *cb;
+	unsigned char ch;
+
+	cb = tty->cooked_q.head;
+	col = count = 0;
+
+	while(cb) {
+		for(n = 0; n < cb->end_off; n++) {
+			if(n >= cb->start_off) {
+				ch = cb->data[n];
+				if(ch == '\t') {
+					while(!tty->tab_stop[++col]);
+				} else {
+					col++;
+					if(ISCNTRL(ch) && !ISSPACE(ch) && tty->termios.c_lflag & ECHOCTL) {
+						col++;
+					}
+				}
+				col %= tty->winsize.ws_col;
+			}
+		}
+		cb = cb->next;
+	}
+	count = tty->column - col;
+
+	while(count--) {
+		charq_putchar(&tty->write_q, '\b');
+		tty->column--;
+	}
+}
+
 void do_cook(struct tty *tty)
 {
 	int n;
@@ -373,15 +471,21 @@ void do_cook(struct tty *tty)
 
 			if(tty->termios.c_iflag & IXON) {
 				if(ch == tty->termios.c_cc[VSTART]) {
-					tty->start(tty);
+					if(tty->start) {
+						tty->start(tty);
+					}
 					continue;
 				}
 				if(ch == tty->termios.c_cc[VSTOP]) {
-					tty->stop(tty);
+					if(tty->stop) {
+						tty->stop(tty);
+					}
 					continue;
 				}
 				if(tty->termios.c_iflag & IXANY) {
-					tty->start(tty);
+					if(tty->start) {
+						tty->start(tty);
+					}
 				}
 			}
 		}
@@ -409,37 +513,42 @@ void do_cook(struct tty *tty)
 		charq_putchar(&tty->cooked_q, ch);
 		tty->flags &= ~TTY_HAS_LNEXT;
 	}
-	tty->output(tty);
+	if(tty->output) {
+		tty->output(tty);
+	}
 	if(!(tty->termios.c_lflag & ICANON) || ((tty->termios.c_lflag & ICANON) && tty->canon_data)) {
 		wakeup(&do_select);
 	}
 	wakeup(&tty->read_q);
 }
 
-int tty_open(struct inode *i, struct fd *fd_table)
+int tty_open(struct inode *i, struct fd *f)
 {
 	int noctty_flag;
-	struct tty *tty;
+	struct tty *tty, *otty;
+	struct inode *oi;
+	struct devpts_files *dp;
 	int errno;
 	 
-	noctty_flag = fd_table->flags & O_NOCTTY;
+	noctty_flag = f->flags & O_NOCTTY;
 
-	if(MAJOR(i->rdev) == SYSCON_MAJOR && MINOR(i->rdev) == 0) {
+	if(i->rdev == TTY_DEV) {
 		if(!current->ctty) {
 			return -ENXIO;
 		}
 	}
 
-	if(MAJOR(i->rdev) == VCONSOLES_MAJOR && MINOR(i->rdev) == 0) {
+	if(i->rdev == VCONSOLE_DEV) {
 		noctty_flag = 1;
 	}
 
 	if(!(tty = get_tty(i->rdev))) {
 		printk("%s(): oops! (%x)\n", __FUNCTION__, i->rdev);
-		printk("_syscondev = %x\n", kparm_syscondev);
+		printk("_syscondev = %x\n", kparms.syscondev);
 		return -ENXIO;
 	}
 
+	errno = 0;
 	if(tty->open) {
 		if((errno = tty->open(tty)) < 0) {
 			return errno;
@@ -448,6 +557,35 @@ int tty_open(struct inode *i, struct fd *fd_table)
 	tty->count++;
 	tty->column = 0;
 
+	f->private_data = tty;
+
+#ifdef CONFIG_UNIX98_PTYS
+	if(i->rdev == PTMX_DEV) {
+		dp = (struct devpts_files *)tty->driver_data;
+		oi = (struct inode *)dp->inode;
+		if((otty = register_tty(oi->rdev))) {
+			otty->count++;
+			otty->driver_data = tty->driver_data;
+			otty->flags |= TTY_PTY_LOCK;
+			otty->link = tty;
+			/* FIXME
+			otty->stop =
+			otty->start =
+			*/
+			otty->deltab = tty_deltab;
+			otty->input = do_cook;
+			otty->output = pty_wakeup_read;
+			otty->open = pty_open;
+			otty->close = pty_close;
+			otty->ioctl = pty_ioctl;
+			f->private_data = otty;
+		} else {
+			printk("WARNING: %s(): unable to register pty slave (%d,%d).\n", __FUNCTION__, MAJOR(oi->rdev), MINOR(oi->rdev));
+			return -ENOMEM;
+		}
+	}
+#endif /* CONFIG_UNIX98_PTYS */
+
 	if(SESS_LEADER(current) && !current->ctty && !noctty_flag && !tty->sid) {
 		current->ctty = tty;
 		tty->sid = current->sid;
@@ -456,17 +594,13 @@ int tty_open(struct inode *i, struct fd *fd_table)
 	return 0;
 }
 
-int tty_close(struct inode *i, struct fd *fd_table)
+int tty_close(struct inode *i, struct fd *f)
 {
 	struct proc *p;
 	struct tty *tty;
 	int errno;
 
-	if(!(tty = get_tty(i->rdev))) {
-		printk("%s(): oops! (%x)\n", __FUNCTION__, i->rdev);
-		return -ENXIO;
-	}
-
+	tty = f->private_data;
 	if(tty->close) {
 		if((errno = tty->close(tty)) < 0) {
 			return errno;
@@ -488,7 +622,7 @@ int tty_close(struct inode *i, struct fd *fd_table)
 	return 0;
 }
 
-int tty_read(struct inode *i, struct fd *fd_table, char *buffer, __size_t count)
+int tty_read(struct inode *i, struct fd *f, char *buffer, __size_t count)
 {
 	unsigned int min;
 	unsigned char ch;
@@ -496,10 +630,7 @@ int tty_read(struct inode *i, struct fd *fd_table, char *buffer, __size_t count)
 	struct callout_req creq;
 	int n;
 
-	if(!(tty = get_tty(i->rdev))) {
-		printk("%s(): oops! (%x)\n", __FUNCTION__, i->rdev);
-		return -ENXIO;
-	}
+	tty = f->private_data;
 
 	/* only the foreground process group is allowed to read from the tty */
 	if(current->ctty == tty && current->pgid != tty->pgid) {
@@ -525,6 +656,10 @@ int tty_read(struct inode *i, struct fd *fd_table, char *buffer, __size_t count)
 				break;
 			}
 		}
+		if(tty->flags & TTY_OTHER_CLOSED) {
+			n = -EIO;
+			break;
+		}
 
 		if(tty->termios.c_lflag & ICANON) {
 			if((ch = LAST_CHAR(&tty->cooked_q))) {
@@ -559,7 +694,7 @@ int tty_read(struct inode *i, struct fd *fd_table, char *buffer, __size_t count)
 						creq.fn = wait_vtime_off;
 						creq.arg = (unsigned int)&tty->cooked_q;
 						add_callout(&creq, timeout);
-						if(fd_table->flags & O_NONBLOCK) {
+						if(f->flags & O_NONBLOCK) {
 							return -EAGAIN;
 						}
 						if(sleep(&tty->read_q, PROC_INTERRUPTIBLE)) {
@@ -588,7 +723,7 @@ int tty_read(struct inode *i, struct fd *fd_table, char *buffer, __size_t count)
 						creq.fn = wait_vtime_off;
 						creq.arg = (unsigned int)&tty->cooked_q;
 						add_callout(&creq, timeout);
-						if(fd_table->flags & O_NONBLOCK) {
+						if(f->flags & O_NONBLOCK) {
 							n = -EAGAIN;
 							break;
 						}
@@ -603,24 +738,24 @@ int tty_read(struct inode *i, struct fd *fd_table, char *buffer, __size_t count)
 					}
 				}
 			} else {
-				if(tty->cooked_q.count > 0) {
-					if(min < tty->termios.c_cc[VMIN] || !tty->termios.c_cc[VMIN]) {
-						if(n < count) {
-							ch = charq_getchar(&tty->cooked_q);
-							buffer[n++] = ch;
-							if(--tty->canon_data < 0) {
-								tty->canon_data = 0;
-							}
+				while(tty->cooked_q.count > 0) {
+					if(n < count) {
+						ch = charq_getchar(&tty->cooked_q);
+						buffer[n++] = ch;
+						if(--tty->canon_data < 0) {
+							tty->canon_data = 0;
 						}
-						min++;
+					} else {
+						break;
 					}
+					min++;
 				}
 				if(min >= tty->termios.c_cc[VMIN]) {
 					break;
 				}
 			}
 		}
-		if(fd_table->flags & O_NONBLOCK) {
+		if(f->flags & O_NONBLOCK) {
 			n = -EAGAIN;
 			break;
 		}
@@ -636,16 +771,13 @@ int tty_read(struct inode *i, struct fd *fd_table, char *buffer, __size_t count)
 	return n;
 }
 
-int tty_write(struct inode *i, struct fd *fd_table, const char *buffer, __size_t count)
+int tty_write(struct inode *i, struct fd *f, const char *buffer, __size_t count)
 {
 	unsigned char ch;
 	struct tty *tty;
 	int n;
 
-	if(!(tty = get_tty(i->rdev))) {
-		printk("%s(): oops! (%x)\n", __FUNCTION__, i->rdev);
-		return -ENXIO;
-	}
+	tty = f->private_data;
 
 	/* only the foreground process group is allowed to write to the tty */
 	if(current->ctty == tty && current->pgid != tty->pgid) {
@@ -673,11 +805,13 @@ int tty_write(struct inode *i, struct fd *fd_table, const char *buffer, __size_t
 			}
 			n++;
 		}
-		tty->output(tty);
+		if(tty->output) {
+			tty->output(tty);
+		}
 		if(n == count) {
 			break;
 		}
-		if(fd_table->flags & O_NONBLOCK) {
+		if(f->flags & O_NONBLOCK) {
 			n = -EAGAIN;
 			break;
 		}
@@ -699,16 +833,14 @@ int tty_write(struct inode *i, struct fd *fd_table, const char *buffer, __size_t
 }
 
 /* FIXME: http://www.lafn.org/~dave/linux/termios.txt (doc/termios.txt) */
-int tty_ioctl(struct inode *i, int cmd, unsigned int arg)
+int tty_ioctl(struct inode *i, struct fd *f, int cmd, unsigned int arg)
 {
 	struct proc *p;
 	struct tty *tty;
 	int errno;
 
-	if(!(tty = get_tty(i->rdev))) {
-		printk("%s(): oops! (%x)\n", __FUNCTION__, i->rdev);
-		return -ENXIO;
-	}
+	tty = f->private_data;
+	errno = 0;
 
 	switch(cmd) {
 		/*
@@ -833,10 +965,14 @@ int tty_ioctl(struct inode *i, int cmd, unsigned int arg)
 		case TCXONC:
 			switch(arg) {
 				case TCOOFF:
-					tty->stop(tty);
+					if(tty->stop) {
+						tty->stop(tty);
+					}
 					break;
 				case TCOON:
-					tty->start(tty);
+					if(tty->start) {
+						tty->start(tty);
+					}
 					break;
 				default:
 					return -EINVAL;
@@ -939,16 +1075,10 @@ int tty_ioctl(struct inode *i, int cmd, unsigned int arg)
 				return errno;
 			}
 			changed = 0;
-			if(tty->winsize.ws_row != ws->ws_row) {
-				changed = 1;
-			}
-			if(tty->winsize.ws_col != ws->ws_col) {
-				changed = 1;
-			}
-			if(tty->winsize.ws_xpixel != ws->ws_xpixel) {
-				changed = 1;
-			}
-			if(tty->winsize.ws_ypixel != ws->ws_ypixel) {
+			if(tty->winsize.ws_row != ws->ws_row ||
+			   tty->winsize.ws_col != ws->ws_col ||
+			   tty->winsize.ws_xpixel != ws->ws_xpixel ||
+			   tty->winsize.ws_ypixel != ws->ws_ypixel) {
 				changed = 1;
 			}
 			tty->winsize.ws_row = ws->ws_row;
@@ -984,11 +1114,23 @@ int tty_ioctl(struct inode *i, int cmd, unsigned int arg)
 			}
 			break;
 		}
-
+		case TIOCINQ:
+		{
+			int *val = (int *)arg;
+			if(tty->termios.c_lflag & ICANON) {
+				*val = tty->cooked_q.count;
+			} else {
+				*val = tty->read_q.count;
+			}
+			break;
+		}
 		default:
-			return vt_ioctl(tty, cmd, arg);
+			errno = vt_ioctl(tty, cmd, arg);
+			if(tty->ioctl) {
+				errno = tty->ioctl(tty, f, cmd, arg);
+			}
 	}
-	return 0;
+	return errno;
 }
 
 __loff_t tty_llseek(struct inode *i, __loff_t offset)
@@ -996,14 +1138,11 @@ __loff_t tty_llseek(struct inode *i, __loff_t offset)
 	return -ESPIPE;
 }
 
-int tty_select(struct inode *i, int flag)
+int tty_select(struct inode *i, struct fd *f, int flag)
 {
 	struct tty *tty;
 
-	if(!(tty = get_tty(i->rdev))) {
-		printk("%s(): oops! (%x)\n", __FUNCTION__, i->rdev);
-		return 0;
-	}
+	tty = f->private_data;
 
 	switch(flag) {
 		case SEL_R:
@@ -1028,5 +1167,5 @@ int tty_select(struct inode *i, int flag)
 void tty_init(void)
 {
 	charq_init();
-	memset_b(tty_table, 0, sizeof(tty_table));
+	tty_table = NULL;
 }
