@@ -19,34 +19,19 @@
 #include <fiwix/stat.h>
 #include <fiwix/blk_queue.h>
 
-#define BUFFER_HASH(dev, block)	(((__dev_t)(dev) ^ (__blk_t)(block)) % (NR_BUF_HASH))
 #define NR_BUF_HASH		(buffer_hash_table_size / sizeof(unsigned int))
+#define BUFFER_HASH(dev, block)	(((__dev_t)(dev) ^ (__blk_t)(block)) % (NR_BUF_HASH))
 #define BUFHEAD_INDEX(size)	((size / BLKSIZE_1K) - 1)
 
 #define NO_GROW		0
 #define GROW_IF_NEEDED	1
 
 struct buffer *buffer_table;		/* buffer pool */
+struct buffer **buffer_hash_table;
 
 /* [0] = 1KB, [1] = 2KB, [2] = unused, [3] = 4KB */
 struct buffer *buffer_head[4];		/* heads of free list */
 struct buffer *buffer_dirty_head[4];	/* heads of dirty list */
-struct buffer *buffer_retained_head[4];	/* heads of retained list */
-
-/*
- * hash table
- * +--------+  +--------------+  +--------------+  +--------------+
- * | index  |  |prev|data|next|  |prev|data|next|  |prev|data|next|
- * |   0   --> | /  |    | --->  <--- |    | --->  <--- |    |  / |
- * +--------+  +--------------+  +--------------+  +--------------+
- * +--------+  +--------------+  +--------------+  +--------------+
- * | index  |  |prev|data|next|  |prev|data|next|  |prev|data|next|
- * |   1   --> | /  |    | --->  <--- |    | --->  <--- |    |  / |
- * +--------+  +--------------+  +--------------+  +--------------+
- *              (buffer)          (buffer)          (buffer)
- *    ...
- */
-struct buffer **buffer_hash_table;
 
 static struct resource sync_resource = { 0, 0 };
 
@@ -233,6 +218,24 @@ static void insert_on_free_list(struct buffer *buf)
 	h->prev_free = buf;
 }
 
+static void append_on_free_list(struct buffer *buf)
+{
+	struct buffer *h;
+	int index;
+
+	index = BUFHEAD_INDEX(buf->size);
+	h = buffer_head[index];
+
+	if(!h) {
+		buffer_head[index] = buf;
+		h = buffer_head[index];
+	} else {
+		buf->prev_free = h->prev_free;
+		h->prev_free->next_free = buf;
+	}
+	h->prev_free = buf;
+}
+
 static void remove_from_free_list(struct buffer *buf)
 {
 	struct buffer *h;
@@ -323,35 +326,34 @@ static struct buffer *get_free_buffer(int mode, int size)
 {
 	unsigned int flags;
 	struct buffer *buf;
-	int index;
+	int index, min;
 
 	index = BUFHEAD_INDEX(size);
 	buf = buffer_head[index];
 
 	/*
 	 * We check buf->dev to see if this buffer has been already used
-	 * and, if that's the case, then we know that there aren't more
-	 * unused buffers, so let's try to create new buffers instead of
-	 * reusing them.
+	 * and, if so, then we know that there aren't more unused buffers,
+	 * so let's try to create new ones instead of reusing them.
 	 */
 	if(!buf || (buf && buf->dev)) {
 		if(mode == GROW_IF_NEEDED) {
-			if(kstat.free_pages > (kstat.min_free_pages + 1)) {
+			min = (kstat.total_mem_pages * FREE_PAGES_RATIO) / 100;
+			if(kstat.free_pages > min) {
 				if((buf = create_buffers(size))) {
 					return buf;
 				}
 			}
-			buf = buffer_head[index];
 		}
 	}
-	if(!buf) {
-		/* no more buffers in this free list */
-		return NULL;
-	}
 
 	for(;;) {
 		SAVE_FLAGS(flags); CLI();
-		buf = buffer_head[index];
+		if(!(buf = buffer_head[index])) {
+			/* no more buffers in this free list */
+			RESTORE_FLAGS(flags);
+			return NULL;
+		}
 		if(buf->flags & BUFFER_LOCKED) {
 			sleep(&buffer_wait, PROC_UNINTERRUPTIBLE);
 		} else {
@@ -532,7 +534,6 @@ struct buffer *bread(__dev_t dev, __blk_t block, int size)
 			return buf;
 		}
 		if(!(d = get_device(BLK_DEV, dev))) {
-			printk("WARNING: %s(): device major %d not found!\n", __FUNCTION__, MAJOR(dev));
 			return NULL;
 		}
 		if(do_blk_request(d, d->fsop->read_block, buf) == size) {
@@ -541,7 +542,6 @@ struct buffer *bread(__dev_t dev, __blk_t block, int size)
 		}
 		brelse(buf);
 	}
-	
 	printk("WARNING: %s(): returning NULL!\n", __FUNCTION__);
 	return NULL;
 }
@@ -587,7 +587,7 @@ void sync_buffers(__dev_t dev)
 			if(first == buf) {
 				insert_on_dirty_list(buf);
 				buf->flags &= ~BUFFER_LOCKED;
-				wakeup(&buffer_wait);
+				flushed = 1;
 				break;
 			}
 			if(!dev || buf->dev == dev) {
@@ -596,7 +596,7 @@ void sync_buffers(__dev_t dev)
 					buf->flags &= ~BUFFER_LOCKED;
 					continue;
 				}
-				flushed++;
+				flushed = 1;
 			} else {
 				if(!first) {
 					first = buf;
@@ -637,35 +637,38 @@ void invalidate_buffers(__dev_t dev)
 static int reclaim_siblings(struct buffer *buf)
 {
 	struct buffer *orig, *tmp;
-	int index;
+	unsigned int flags;
 
-	index = BUFHEAD_INDEX(buf->size);
 	orig = buf;
 
 	if(buf->first_sibling) {
 		buf = buf->first_sibling;
 	}
-	/* check if one of the siblings is locked or dirty */
+	/* abort if one of the siblings is locked */
 	do {
-		if(buf == orig) {
-			buf = buf->next_sibling;
-			continue;
-		}
-		if(buf->flags & (BUFFER_LOCKED | BUFFER_DIRTY)) {
-			/*
-			 * If one of the siblings is not eligible to be freed up, then
-			 * we give up and return without brelse(orig), otherwise
-			 * get_free_buffer() will return the same buffer again. So we
-			 * must retain 'orig' in a temporary list to be brelse()d later.
-			 */
-			orig->next_retained = buffer_retained_head[index];
-			buffer_retained_head[index] = orig;
-			return 1;
+		if(buf != orig) {
+			if(buf->flags & BUFFER_LOCKED) {
+				return 1;
+			}
 		}
 		buf = buf->next_sibling;
 	} while(buf);
 
-	/* OK, all siblings are eligible to be freed up */
+	/* OK, all siblings are eligible to be freed up, let's lock them all */
+	SAVE_FLAGS(flags); CLI();
+	buf = orig;
+	if(buf->first_sibling) {
+		buf = buf->first_sibling;
+	}
+	do {
+		if(buf != orig) {
+			buf->flags |= BUFFER_LOCKED;
+		}
+		buf = buf->next_sibling;
+	} while(buf);
+	RESTORE_FLAGS(flags);
+
+	/* now free them up */
 	buf = orig;
 	if(buf->first_sibling) {
 		buf = buf->first_sibling;
@@ -675,6 +678,13 @@ static int reclaim_siblings(struct buffer *buf)
 			buf = buf->next_sibling;
 			continue;
 		}
+		if(buf->flags & BUFFER_DIRTY) {
+			if(!sync_one_buffer(buf)) {
+				remove_from_dirty_list(buf);
+			} else {
+				/* FIXME: undo all locks and give up */
+			}
+		}
 		tmp = buf;
 		buf = buf->next_sibling;
 		remove_from_hash(tmp);
@@ -691,23 +701,26 @@ static int reclaim_siblings(struct buffer *buf)
  */
 int reclaim_buffers(void)
 {
-	struct buffer *buf, *tmp, *retained;
-	int size, found, reclaimed, index;
+	struct buffer *buf;
+	int size, found, reclaimed;
+	unsigned int flags, mark;
 
 	found = reclaimed = 0;
 	size = BLKSIZE_1K;
+	mark = kstat.uptime;
 
 	/* iterate through all buffer sizes */
 	STI();
 	for(;;) {
-		if(size > PAGE_SIZE) {
-			if(!found) {
-				break;
-			}
-			size = BLKSIZE_1K;
-			found = 0;
-		}
 		if((buf = get_free_buffer(NO_GROW, size))) {
+			if(buf->mark == mark) {
+				SAVE_FLAGS(flags); CLI();
+				buf->flags &= ~BUFFER_LOCKED;
+				buf->mark = 0;
+				append_on_free_list(buf);
+				RESTORE_FLAGS(flags);
+				goto next;
+			}
 			found++;
 			if(buf->flags & BUFFER_DIRTY) {
 				if(!sync_one_buffer(buf)) {
@@ -715,30 +728,37 @@ int reclaim_buffers(void)
 				}
 			}
 			if(reclaim_siblings(buf)) {
+				/*
+				 * If one of the siblings is not eligible to be
+				 * freed up, then we release this buffer without
+				 * using brelse(), otherwise get_free_buffer()
+				 * will return the same buffer again.
+				 */
+				SAVE_FLAGS(flags); CLI();
+				buf->flags &= ~BUFFER_LOCKED;
+				buf->mark = mark;
+				append_on_free_list(buf);
+				RESTORE_FLAGS(flags);
 				continue;
 			}
 			kfree((unsigned int)(buf->data) & PAGE_MASK);
 			remove_from_hash(buf);
 			kstat.buffers_size -= buf->size / 1024;
 			del_buffer_from_pool(buf);
-			reclaimed++;
-		}
-		if(reclaimed == NR_BUF_RECLAIM) {
-			break;
+			if(++reclaimed == NR_BUF_RECLAIM) {
+				break;
+			}
+			continue;
 		}
+next:
 		size <<= 1;
-	}
-
-	/* release all retained buffers */
-	for(size = BLKSIZE_1K; size <= PAGE_SIZE; size <<= 1) {
-		index = BUFHEAD_INDEX(size);
-		retained = buffer_retained_head[index];
-		while(retained) {
-			tmp = retained;
-			retained = retained->next_retained;
-			brelse(tmp);
+		if(size > PAGE_SIZE) {
+			if(!found) {
+				break;
+			}
+			size = BLKSIZE_1K;
+			found = 0;
 		}
-		buffer_retained_head[index] = NULL;
 	}
 
 	wakeup(&get_free_buffer);
@@ -748,13 +768,9 @@ int reclaim_buffers(void)
 	 * If some buffers were reclaimed, then wakeup any process
 	 * waiting for a new page because release_page() won't do it.
 	 */
-	if(reclaimed && reclaimed <= NR_BUF_RECLAIM) {
+	if(reclaimed) {
 		wakeup(&get_free_page);
 	}
-
-	if(!reclaimed) {
-		printk("WARNING: %s(): no more buffers on free lists!\n", __FUNCTION__);
-	}
 	return reclaimed;
 }
 
@@ -806,6 +822,7 @@ int kbdflushd(void)
 						break;
 					}
 					flushed = 0;
+					wakeup(&buffer_wait);
 					do_sched();
 				}
 			}
@@ -819,7 +836,6 @@ void buffer_init(void)
 	buffer_table = NULL;
 	memset_b(buffer_head, 0, sizeof(buffer_head));
 	memset_b(buffer_dirty_head, 0, sizeof(buffer_dirty_head));
-	memset_b(buffer_retained_head, 0, sizeof(buffer_retained_head));
 	kstat.max_dirty_buffers = (kstat.max_buffers_size * BUFFER_DIRTY_RATIO) / 100;
 	memset_b(buffer_hash_table, 0, buffer_hash_table_size);
 }
